const Applet = imports.ui.applet;
const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;
const AppFavorites = imports.ui.appFavorites;
const Tweener = imports.ui.tweener;
const DND = imports.ui.dnd;
const Settings = imports.ui.settings;
const AppletDir = imports.ui.appletManager.applets['menu@cdos.org'];

const Cinnamon = imports.gi.Cinnamon;
const St = imports.gi.St;
const Clutter = imports.gi.Clutter;
const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;
const Gtk = imports.gi.Gtk;
const Meta = imports.gi.Meta;
const AccountsService = imports.gi.AccountsService;
const Pango = imports.gi.Pango;

var APIMenu;
try {
    APIMenu = imports.gi.CMenu;
} catch (e) {
    APIMenu = imports.gi.GMenu;
}
const CMenu = APIMenu;

const GnomeSession = imports.misc.gnomeSession;
const DocInfo = imports.misc.docInfo; 
const ScreenSaver = imports.misc.screenSaver;
const Util = imports.misc.util;
const FileUtils = imports.misc.fileUtils;
const Pinyin = imports.misc.pinyin;

const Mainloop = imports.mainloop;
const Lang = imports.lang;
const Cairo = imports.cairo; //
const Gettext = imports.gettext;
const Signals = imports.signals;

const ICON_SIZE = 16;
const MAX_FAV_ICON_SIZE = 36;
const CATEGORY_ICON_SIZE = 22;
const APPLICATION_ICON_SIZE = 18;
const HOVER_ICON_SIZE = 62;
const MAX_RECENT_FILES = 4;
const IMAGE_ICON_SIZE = 10;
const USER_DESKTOP_PATH = FileUtils.getUserDesktopDir();

let appsys = Cinnamon.AppSystem.get_default();
let pinyin = Pinyin.pinyin;
let menuManager; //定义全局菜单管理器变量
let shutdownMenu; //定义全局关机菜单变量
let MyAppletMenu; //
let ori;
let GtkRecent = new Gtk.RecentManager();

var Session = new GnomeSession.SessionManager();

function _(str) {
    let resultConf = Gettext.dgettext("menu@cdos.org", str);
    if (resultConf != str) {
        return resultConf;
    }
    return Gettext.gettext(str);
}

function ApplicationContextMenuItem(appButton, label, action) {
    this._init(appButton, label, action);
}

ApplicationContextMenuItem.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(appButton, label, action) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            focusOnHover: false
        });

        this._appButton = appButton;
        this._action = action;
        this.label = new St.Label({
            text: label
        });
        this.addActor(this.label);
    },

    activate: function(event) {
        this._appButton.toggleMenu();
        let settings = new Gio.Settings({
            schema: 'org.cdos'
        });
        let desktopFiles;
        switch (this._action) {
            case "add_to_panel": //添加到面板的实现
                desktopFiles = settings.get_strv('panel-launchers');
                if(desktopFiles.length < 8){
                    desktopFiles.push(this._appButton.app.get_id());
                    settings.set_strv('panel-launchers', desktopFiles);
                }else{
                    Main.notify(_('Window List'), _('too many favorite applications, please delete some first'));
                }
                break;
            case "remove_from_panel":
                desktopFiles = settings.get_strv('panel-launchers');
                let idx = desktopFiles.indexOf(this._appButton.app.get_id());
                desktopFiles.splice(idx,1);
                settings.set_strv('panel-launchers', desktopFiles);
                break;
            case "add_to_desktop": //添加到桌面的实现
                let file = Gio.file_new_for_path(this._appButton.app.get_app_info().get_filename());
                let destFile = Gio.file_new_for_path(USER_DESKTOP_PATH + "/" + this._appButton.app.get_id());
                try {
                    file.copy(destFile, 0, null, function() {});
                    // Need to find a way to do that using the Gio library, but modifying the access::can-execute attribute on the file object seems unsupported
                    Util.spawnCommandLine("chmod +x \"" + USER_DESKTOP_PATH + "/" + this._appButton.app.get_id() + "\"");
                } catch (e) {
                    global.log(e);
                }
                break;
            case "add_to_favorites": //添加到收藏夹的实现
                AppFavorites.getAppFavorites().addFavorite(this._appButton.app.get_id());
                break;
            case "remove_from_favorites": //从收藏中删除的实现
                AppFavorites.getAppFavorites().removeFavorite(this._appButton.app.get_id());
                break;
            case "remove_recent_all":
               //GtkRecent = new Gtk.RecentManager();
                GtkRecent.purge_items();
                break;
            case "remove_recent_item":
                GtkRecent.remove_item(this._appButton.file.uri);
        }
        return false;
    }
};

function GenericApplicationButton(appsMenuButton, app) {
    this._init(appsMenuButton, app);
}

GenericApplicationButton.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(appsMenuButton, app, withMenu) {
        this.app = app;
        this.appsMenuButton = appsMenuButton;
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false,
            focusOnHover: false
        });

        this.withMenu = withMenu;

        if (this.withMenu) {
            this.menu = new Contextmenu(this);
            this.menu.actor.set_style_class_name('menu-context-menu');
            MyAppletMenu.addChildMenu(this.menu);
        }

        return true;
    },

    _onButtonReleaseEvent: function(actor, event) {
        if (event.get_button() == 1) {
            if(this.menu.isOpen){
                this.menu.close();
                return;
            }
            this.activate(event);
        }
        if (event.get_button() == 2) {
            //this.appsMenuButton._clearSelections(this.appsMenuButton.categoriesBox);
        }
        if (event.get_button() == 3) {
            let pos = event.get_coords();
            this.appsMenuButton.closeApplicationsContextMenus(this.app, true);
            this.menu.x = pos[0] + 2;
            this.menu.y = pos[1] + 2;
            this.toggleMenu();
        }
    },

    setActive: function () {
        return true;
    },

    activate: function(event) {
        this.app.open_new_window(-1);
        this.appsMenuButton.menu.close();
    },

    closeMenu: function() {
        if (this.withMenu) this.menu.close();
    },

    toggleMenu: function() {
        if (!this.withMenu) return;

        if (!this.menu.isOpen) {
            let children = this.menu.box.get_children();
            for (var i in children) {
                children[i].destroy();
            }
            let settings = new Gio.Settings({
                schema: 'org.cdos'
            });
            let desktopFiles = settings.get_strv('panel-launchers');

            let menuItem;
            if(desktopFiles.indexOf(this.app.get_id()) == -1){
                menuItem = new ApplicationContextMenuItem(this, _("Add to panel"), "add_to_panel");
                this.menu.addMenuItem(menuItem);
            }else{
                menuItem = new ApplicationContextMenuItem(this, _("Remove from panel"), "remove_from_panel");
                this.menu.addMenuItem(menuItem);
            }

            if (USER_DESKTOP_PATH) {
                menuItem = new ApplicationContextMenuItem(this, _("Add to desktop"), "add_to_desktop");
                this.menu.addMenuItem(menuItem);
            }
            if (AppFavorites.getAppFavorites().isFavorite(this.app.get_id())) {
                menuItem = new ApplicationContextMenuItem(this, _("Remove from quick operation"), "remove_from_favorites");
                this.menu.addMenuItem(menuItem);
            } else {
                menuItem = new ApplicationContextMenuItem(this, _("Add to quick operation"), "add_to_favorites");
                this.menu.addMenuItem(menuItem);
            }
        }
        this.menu.toggle();
    },
}

function ApplicationButton(appsMenuButton, app) {
    this._init(appsMenuButton, app);
}

ApplicationButton.prototype = {
    __proto__: GenericApplicationButton.prototype,

    _init: function(appsMenuButton, app) {
        GenericApplicationButton.prototype._init.call(this, appsMenuButton, app, true);

        this.actor.add_style_class_name('menu-application-button');
        this.icon = this.app.create_icon_texture(APPLICATION_ICON_SIZE);
        this.label = new St.Label({
            text: this.app.get_name(),
            style_class: 'menu-application-button-label'
        });
        this.addActor(this.icon);
        this.addActor(this.label);

        //this._draggable = DND.makeDraggable(this.actor,{dragActorOpacity: 200});
        //this.isDraggableApp = true;
    },

    get_app_id: function() {
        return this.app.get_id();
    },

    getDragActor: function() {
        let favorites = AppFavorites.getAppFavorites().getFavorites();
        let nbFavorites = favorites.length;
        let monitorHeight = Main.layoutManager.primaryMonitor.height;
        let real_size = (0.7 * monitorHeight) / nbFavorites;
        let icon_size = 0.6 * real_size;
        if (icon_size > MAX_FAV_ICON_SIZE) icon_size = MAX_FAV_ICON_SIZE;
        return this.app.create_icon_texture(icon_size);
    },

    // Returns the original actor that should align with the actor
    // we show as the item is being dragged.
    getDragActorSource: function() {
        return this.actor;
    }
};
Signals.addSignalMethods(ApplicationButton.prototype);

function PlaceButton(appsMenuButton, place, button_name) {
    this._init(appsMenuButton, place, button_name);
}

PlaceButton.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(appsMenuButton, place, button_name) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        this.actor.add_style_class_name('menu-category-button');
        this.place = place;
        this.button_name = button_name;
        this.label = new St.Label({
            text: this.button_name,
            style_class: 'menu-application-button-label'
        });
        this.icon = place.iconFactory(APPLICATION_ICON_SIZE);
        this.addActor(this.icon);
        this.addActor(this.label);
    },

    activate: function() {
        this.place.launch();
    },

    _onButtonReleaseEvent: function(actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(PlaceButton.prototype);

function ApplicationMenu() {
    this._init.apply(this, arguments);
}

ApplicationMenu.prototype = {
    __proto__: PopupMenu.PopupMenuBase.prototype,

    _init: function(sourceActor) {
        PopupMenu.PopupMenuBase.prototype._init.call(this, sourceActor);

        this.actor = new St.BoxLayout({
            style_class: 'popup-sub-menu',
            vertical: true
        });

        this.actor.add_actor(this.box);
        this.actor._delegate = this;
        this.actor.clip_to_allocation = true;
        //this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
        this.actor.hide();
    },

    toggle: function () {

    },
    open: function () {

    },
    close: function () {
        //this.closex();
    },

    togglex: function() {
        let focus = global.stage.key_focus;
        //log(focus);
        if (this.isOpen)
            this.closex();
        else
            this.openx();
    },
    openx: function(animate) {
        if (this.isOpen)
            return;

        this.isOpen = true;

        this.actor.show();

        if (animate) {
            let [minHeight, naturalHeight] = this.actor.get_preferred_height(-1);
            this.actor.height = 0;
            Tweener.addTween(this.actor, {
                height: naturalHeight,
                time: 0.25, //menu-context-menu打开时间
                //time: 5,   //change
                onCompleteScope: this,
                onComplete: function() {
                    this.actor.set_height(-1);
                    this.emit('open-state-changed', true);
                }
            });
        } else {
            this.emit('open-state-changed', true);
        }
    },

    closex: function(animate) {
        if (!this.isOpen)
            return;
        this.isOpen = false;

        if (animate) {
            Tweener.addTween(this.actor, {
                height: 0,
                time: 0.25, //menu-context-menu关闭时间
                //time: 5,  //change
                onCompleteScope: this,
                onComplete: function() {
                    this.actor.hide();
                    this.actor.set_height(-1);
                    this.emit('open-state-changed', false);
                },
            });
        } else {
            this.actor.hide();

            this.isOpen = false;
            this.emit('open-state-changed', false);
        }
    }
};

/*
 * 类目
*/
function CategoryButton(category) {
    this._init(category);
}

CategoryButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function(category) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false,
            focusOnHover: true
        });
        var label;
        this.actor.add_style_class_name('menu-category-button');
        if (category) {
            let icon = category.get_icon();
            if (icon && icon.get_names) this.icon_name = icon.get_names().toString();
            else this.icon_name = "";
            label = category.get_name();
        } else label = _("All Applications");
        this.label = new St.Label({
            text: label,
            style_class: 'menu-application-button-label'
        });
        //log(this.icon_name);
        if (category && this.icon_name) {
            this.icon = new St.Icon({
                icon_name: this.icon_name, //'folder-closed',
                icon_size: CATEGORY_ICON_SIZE,
                icon_type: St.IconType.FULLCOLOR
            });
            this.addActor(this.icon);
        }
        this.addActor(this.label);
        this.menu = new ApplicationMenu(this.actor);
        //this.menu.actor.set_style_class_name('menu-context-menu');
        this.menu.connect('open-state-changed', Lang.bind(this, this._subMenuOpenStateChanged));
        this.menu.actor.style = 'padding:1px 0;';
    },

    activate: function() {
        this.menu.togglex();
    },

    removeItems: function() {
        let children = this.menu._getMenuItems();
        for (let i = 0; i < children.length; i++) {
            let item = children[i];
            this.menu.box.remove_actor(item.actor);
            if (item.menu)
                this.menu.box.remove_actor(item.menu.actor);
        }
    },

    destroy: function() {
        this.menu.destroy();
        PopupBaseMenuItem.prototype.destroy.call(this);
    },

    _subMenuOpenStateChanged: function(menu, open) {
        if (open)
            this.actor.add_style_pseudo_class('open');
        else
            this.actor.remove_style_pseudo_class('open');
    },

    _onButtonReleaseEvent: function(actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(CategoryButton.prototype);

function PlaceCategoryButton() {
    this._init();
}

/*
 *位置按钮
*/
PlaceCategoryButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function() {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        this.actor.add_style_class_name('menu-category-button');
        this.label = new St.Label({
            text: _("Places"),
            style_class: 'menu-category-button-label'
        });
        this.icon = new St.Icon({
            icon_name: "folder",
            icon_size: CATEGORY_ICON_SIZE,
            icon_type: St.IconType.FULLCOLOR
        });
        this.addActor(this.icon);
        this.addActor(this.label);
        this.menu = new ApplicationMenu(this.actor);
        this.menu.actor.set_style_class_name('menu-context-menu');
    },

    activate: function() {
        this.menu.toggle();
    },

    removeItems: function() {
        let children = this.menu._getMenuItems();
        for (let i = 0; i < children.length; i++) {
            let item = children[i];
            this.menu.box.remove_actor(item.actor);
            if (item.menu)
                this.menu.box.remove_actor(item.menu.actor);
        }
    },

    destroy: function() {
        this.menu.destroy();
        PopupBaseMenuItem.prototype.destroy.call(this);
    },

    _onButtonReleaseEvent: function(actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(PlaceCategoryButton.prototype);

function FavoritesButton(appsMenuButton, app, nbFavorites) {
    this._init(appsMenuButton, app, nbFavorites);
}

FavoritesButton.prototype = {
    __proto__: GenericApplicationButton.prototype,

    _init: function(appsMenuButton, app, nbFavorites) {
        GenericApplicationButton.prototype._init.call(this, appsMenuButton, app, true);
        let monitorHeight = Main.layoutManager.primaryMonitor.height;
        let icon_size = MAX_FAV_ICON_SIZE;

        this.actor.add_style_class_name('menu-favorites-button');
        this.addActor(app.create_icon_texture(icon_size));

        this.label = new St.Label({
            text: this.app.get_name(),
            style_class: 'menu-favorites-button-label'
        });
        this.addActor(this.label);

        //this._draggable = DND.makeDraggable(this.actor);
        //this.isDraggableApp = true;
    },

    get_app_id: function() {
        return this.app.get_id();
    },

    getDragActor: function() {
        return new Clutter.Clone({
            source: this.actor
        });
    },

    // Returns the original actor that should align with the actor
    // we show as the item is being dragged.
    getDragActorSource: function() {
        return this.actor;
    }
};

/*
 *最近常用按钮
*/
function RecentButton(appsMenuButton, file) {
    this._init(appsMenuButton, file);
}

RecentButton.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(appsMenuButton, file) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });

        let icon_size = MAX_FAV_ICON_SIZE;

        this.appsMenuButton = appsMenuButton;
        this.file = file;

        this.actor.add_style_class_name('menu-favorites-button');
        this.actor._delegate = this;

        let filename = this.file.name;

        this.label = new St.Label({
            text: filename,
            style_class: 'menu-application-button-label'
        });
        this.label.clutter_text.ellipsize = Pango.EllipsizeMode.MIDDLE;
        this.label.set_style('max-width:14em;');

        this.addActor(file.createIcon(icon_size));
        this.addActor(this.label);

        this.menu = new Contextmenu(this);
        this.menu.actor.set_style_class_name('menu-context-menu');
        MyAppletMenu.addChildMenu(this.menu);

        let item = new ApplicationContextMenuItem(this, _("Remove from list"), "remove_recent_item");
        this.menu.addMenuItem(item);
        let item = new ApplicationContextMenuItem(this, _("Clear list"), "remove_recent_all");
        this.menu.addMenuItem(item);
    },

    _onButtonReleaseEvent: function(actor, event) {
        if (event.get_button() == 1) {
            if (this.menu.isOpen) {
                this.menu.close();
                return;
            }
            this.activate(event);
        }else if (event.get_button() == 3) {
            let pos = event.get_coords();
            this.menu.x = pos[0] + 2;
            this.menu.y = pos[1] + 2;
            this.toggleMenu();
        }
    },

    setActive: function () {
        return true;
    },

    activate: function(event) {
        this.file.launch();
        this.appsMenuButton.menu.close();
    },

    toggleMenu: function () {
        this.menu.toggle();
    }
};

function ShortcutItem(label, icon, func, parent, hoverIcon) {
    this._init(label, icon, func, parent, hoverIcon);
}

ShortcutItem.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function(label, icon_name, func, parent, hoverIcon) {
        this.parent = parent;
        this.hoverIcon = hoverIcon;
        this.icon_name = icon_name;

        this.func = func;
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, null);

        this.actor.set_style_class_name('menu-shortcut-item'); //原代码
        let children = this.actor.get_children();
        for(let i =0 ; i < children.length; i++){
            children[i].destroy();
        }

        this.label = new St.Label({
            text: label,
            style_class: 'menu-shortcut-item-label'
        });
        this.addActor(this.label);
    },

    setActive: function(active) {
        if (active) {
            this.actor.remove_style_class_name('menu-shortcut-item');
            this.actor.add_style_class_name('menu-shortcut-item-selected');
            this.hoverIcon._refresh(this.icon_name);

        } else {
            this.actor.remove_style_class_name('menu-shortcut-item-selected');
            this.actor.add_style_class_name('menu-shortcut-item');
            this.hoverIcon._refresh();
        }
    },

    _onButtonReleaseEvent: function(actor, event) {
        if (event.get_button() == 1) {
            this.activate(event);
        }
    },

    activate: function(event) {
        eval(this.func); //执行命令
        this.parent.close(); //关闭主菜单
    }
};

function AllProgramsButton(label, parent) {
    this._init(label, parent);
}

AllProgramsButton.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(label, parent) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            reactive: true
        });

        this.actor.set_style_class_name('menu-allprogram-item menu-category-button');
        this.parent = parent;

        let iconFile;
        let icon;
        //使用图片做为icon
        iconFile = Gio.file_new_for_path(global.datadir + '/theme/icon-002.png');
        icon = new Gio.FileIcon({
            file: iconFile
        });
        this._forwardIcon = St.TextureCache.get_default().load_gicon(null, icon, IMAGE_ICON_SIZE);
        //使用图片做为icon
        iconFile = Gio.file_new_for_path(global.datadir + '/theme/icon-003.png');
        icon = new Gio.FileIcon({
            file: iconFile
        });
        this._backIcon = St.TextureCache.get_default().load_gicon(null, icon, IMAGE_ICON_SIZE);

        this.icon = new St.Bin({
            style_class: 'popup-menu-icon'
        });

        this.icon.set_child(this._backIcon);
        this.label = new St.Label({
            text: label,
            style_class: 'menu-application-button-label'
        });
        this.addActor(this.icon);
        this.addActor(this.label);
    },

    setActive: function(active) {
        if (active)
            this.actor.add_style_class_name('menu-category-button-selected');
        else
            this.actor.remove_style_class_name('menu-category-button-selected');
    },

    _onButtonReleaseEvent: function(actor, event) {
        if (event.get_button() == 1) {
            this.activate(event);
        }
    },

    activate: function(event) {
        if(this.parent.favsBox.visible){
            this.parent.favsBox.hide();
            this.parent.categoriesScrollBox.show();
            this.icon.set_child(this._backIcon);
            this.label.set_text(_("Back"));
            this.parent._appletStyles();
        } else {
            this.parent.favsBox.show();
            let children = this.parent.favsBox.get_children();
            children.forEach(function(child){
                child.emit("style_changed");
            });
            this.parent.categoriesScrollBox.hide();
            this.icon.set_child(this._forwardIcon);
            this.label.set_text(_("All Programs"));
        }
        //this.parent._refreshRecent();
    },
    reset: function () {
        this.icon.set_child(this._forwardIcon);
        this.label.set_text(_("All Programs"));
    }
};

function HoverIcon(menu) {
    this._init(menu);
}

HoverIcon.prototype = {
    _init: function(menu) {
        try {
            this._menu = menu;
            let userBox = new St.BoxLayout({
                style_class: 'user-box',
                reactive: true,
                vertical: true
            }); //change
            this._userIcon = new St.Bin({
                style_class: 'user-icon'
            }); //用户图标
            userBox.connect('button-press-event', Lang.bind(this, function() {
                this._menu.toggle();
                Util.spawnCommandLine("cdos-settings-users");
            }));

            this._userIcon.hide();
            userBox.add(this._userIcon, {
                x_fill: true,
                y_fill: false,
                x_align: St.Align.END,
                y_align: St.Align.START
            });
            this.userLabel = new St.Label(({
                style_class: 'user-label'
            })); //用户名

            let box = new St.BoxLayout({vertical: true});
            box.style = 'padding: 5px 0';
            box.add_actor(this.userLabel);
            userBox.add(box, {
                x_fill: true,
                y_fill: false,
                x_align: St.Align.END,
                y_align: St.Align.END
            }); //change
            this._user = AccountsService.UserManager.get_default().get_user(GLib.get_user_name());
            this._userLoadedId = this._user.connect('notify::is_loaded', Lang.bind(this, this._onUserChanged));
            this._userChangedId = this._user.connect('changed', Lang.bind(this, this._onUserChanged));
            this.actor = new St.Bin({
                style_class: 'menu-hover-icon'
            });
            this.userBox = userBox;     
            this._icon = new St.Icon({
                icon_size: HOVER_ICON_SIZE,
                icon_type: St.IconType.FULLCOLOR
            });
            this.actor.show();
            this.actor.set_child(this.userBox);
            this._onUserChanged();
        } catch (e) {
            global.logError(e);
        }

    },

    _onUserChanged: function() {
        //this._icon.hide();
        this.userBox.show();
        if (this._user.is_loaded) {
            this.userLabel.set_text(GLib.get_user_name() || ' ');
            if (this._userIcon) {
                let iconFileName = this._user.get_icon_file();
                let iconFile = Gio.file_new_for_path(iconFileName);
                let icon;
                if (iconFile.query_exists(null)) {
                    icon = new Gio.FileIcon({
                        file: iconFile
                    });
                } else {
                    icon = new Gio.ThemedIcon({
                        name: 'avatar-default'
                    });
                }
                this.img = St.TextureCache.get_default().load_gicon(null, icon, HOVER_ICON_SIZE);
                this._userIcon.set_child(this.img);
                this._userIcon.add_style_class_name('light');
                this.actor.set_child(this.userBox);
                this._userIcon.show();
                this.userLabel.show();
            }
        }
    },
    _refresh: function(icon_name) {
        if (icon_name) {

            this._userIcon.set_child(this._icon);
            this._userIcon.remove_style_class_name('light');
            this._icon.set_icon_name(icon_name);
            this._icon.show();
            this.userLabel.hide();
        } else {
            this._userIcon.set_child(this.img);
            this._userIcon.add_style_class_name('light');
            this.userBox.show();
            this.userLabel.show();

        }
    }
};

function ShutdownContextMenuItem(parent, appButton, label, action) {
    this._init(parent, appButton, label, action);
}

ShutdownContextMenuItem.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(parent, appButton, label, action) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            focusOnHover: false
        });
        this._session = new GnomeSession.SessionManager();
        this._screenSaverProxy = new ScreenSaver.ScreenSaverProxy();
        this._action = action;
        this.label = new St.Label({
            text: label
        });
        this._appButton = appButton;
        this.parent = parent;
        this.addActor(this.label);
        this.actor.set_style_class_name('shutdown-context-menu-item');
    },

    activate: function(event) {
        this.parent.close();
        this._appButton.menu.toggle();
        switch (this._action) {
            case 'lock-screen':
                let screensaver_settings = new Gio.Settings({
                    schema: "org.cdos.desktop.screensaver"
                });
                let screensaver_dialog = Gio.file_new_for_path("/usr/bin/cdos-screensaver-command");
                if (screensaver_dialog.query_exists(null)) {
                    if (screensaver_settings.get_boolean("ask-for-away-message")) {
                        Util.spawnCommandLine("cdos-screensaver-lock-dialog");
                    } else {
                        Util.spawnCommandLine("cdos-screensaver-command --lock");
                    }
                }
                break;
            case 'switch-user':
                if (GLib.getenv("XDG_SEAT_PATH")) {
                    Util.spawnCommandLine("cdos-screensaver-command --lock");
                    Util.spawnCommandLine("dm-tool switch-to-greeter");
                } else if (GLib.file_test("/usr/bin/mdmflexiserver", GLib.FileTest.EXISTS)) {
                    // MDM
                    Util.spawnCommandLine("mdmflexiserver");
                } else if (GLib.file_test("/usr/bin/gdmflexiserver", GLib.FileTest.EXISTS)) {
                    Util.spawnCommandLine("cdos-screensaver-command --lock");
                    Util.spawnCommandLine("gdmflexiserver");
                }
                break;
            case 'logout':
                this._session.LogoutRemote(0);
                //Util.spawnCommandLine('dbus-send --session --print-reply --dest="org.cdos.SessionManager" /org.cdos/SessionManager org.cdos.SessionManager.RequestLogout');
                break;
            case 'shutdown':
                //this._session.ShutdownRemote();
                Util.spawnCommandLine('dbus-send --session --print-reply --dest="org.cdos.SessionManager" /org.cdos/SessionManager org.cdos.SessionManager.RequestShutdown');
                break;
            case 'reboot':
                Util.spawnCommandLine('dbus-send --session --print-reply --dest="org.cdos.SessionManager" /org.cdos/SessionManager org.cdos.SessionManager.RequestReboot');
                break;
            case 'suspend':
                Util.spawnCommandLine('dbus-send --system --print-reply --dest="org.freedesktop.UPower" /org/freedesktop/UPower org.freedesktop.UPower.Suspend');
                break;
            case 'hibernate':
                Util.spawnCommandLine('dbus-send --system --print-reply --dest="org.freedesktop.UPower" /org/freedesktop/UPower org.freedesktop.UPower.Hibernate');
                break;
        }
        return false;
    }
};


function ShutdownContextMenu(launcher) {
    this._init(launcher);
}
ShutdownContextMenu.prototype = {
    __proto__: PopupMenu.PopupMenu.prototype,
    _init: function(launcher) {
        let orientation = 3;
        AppletPopupMenu.prototype._init.call(this, launcher, orientation);

        this.open = AppletPopupMenu.prototype.open;

        this._arrowSide = 0; //change（自己加的）    useless
        this.setColumnWidths(0); //change（自己加的）  useless
        this.actor.set_opacity(0.1); //change（自己加的） useless
        this.actor.style_class = 'menu-shutdown-menu-box';

        item = new ShutdownContextMenuItem(this, launcher, _("Switch User"), "switch-user"); //切换用户
        this.addMenuItem(item);
        if (GLib.getenv("XDG_SEAT_PATH")) {
            item = new ShutdownContextMenuItem(this, launcher, _("Guest Session"), "guest-session");
            this.addMenuItem(item);
        }

        item = new ShutdownContextMenuItem(this, launcher, _("Log Out..."), "logout"); //注销
        this.addMenuItem(item);

        let item;
        item = new ShutdownContextMenuItem(this, launcher, _("Lock Screen"), "lock-screen"); //锁定屏幕
        this.addMenuItem(item);

        this.addMenuItem(new Separator({style_class: 'contextmenu-separator-line'}));

        let item;
        item = new ShutdownContextMenuItem(this, launcher, _('Reboot'), 'reboot');
        this.addMenuItem(item);

        this.addMenuItem(new Separator({style_class: 'contextmenu-separator-line'}));
        //suspend
        item = new ShutdownContextMenuItem(this, launcher, _("Suspend"), "suspend"); //关机
        this.addMenuItem(item); 
        //hibernate
        item = new ShutdownContextMenuItem(this, launcher, _("Hibernate"), "hibernate"); //关机
        this.addMenuItem(item); 

        this.addMenuItem(new Separator({style_class: 'contextmenu-separator-line'}));

        item = new ShutdownContextMenuItem(this, launcher, _("Power Off..."), "shutdown"); //关机
        this.addMenuItem(item); 
    }
};

Signals.addSignalMethods(ShutdownContextMenu.prototype);

/*
*关机菜单
*/
function ShutdownMenuButton(parent, hoverIcon) { 
    this._init(parent, hoverIcon);
}

ShutdownMenuButton.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(parent, hoverIcon) {
        this._session = new GnomeSession.SessionManager();
        this._screenSaverProxy = new ScreenSaver.ScreenSaverProxy();

        this.hoverIcon = hoverIcon;
        this.parent = parent; //和MyApplet的menu类型一致
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {style_class: 'menu-shutdown-button'});

        let label = new St.Label({text:  _('Shutdown')});
        let labelBin = new St.Bin({x_align: St.Align.MIDDLE, style_class:'menu-shutdown-button-label'});
        labelBin.set_child(label);
        let icon = new St.Bin({style_class: 'menu-shutdown-button-icon'});

        this.addActor(labelBin);
        this.addActor(icon);

        this.menu = new ShutdownContextMenu(this.parent);
        this.menu.connect('before-open', Lang.bind(this,function (menu) {
            let sourceActor = this.actor;
            let sourceNode = sourceActor.get_theme_node();
            let sourceContentBox = sourceNode.get_content_box(sourceActor.get_allocation_box());
            let sourceAllocation = Cinnamon.util_get_transformed_allocation(sourceActor);
            let sourceCenterX = sourceAllocation.x1 + sourceContentBox.x1 + (sourceContentBox.x2 - sourceContentBox.x1);
            let sourceCenterY = sourceAllocation.y1 + sourceContentBox.y1 + (sourceContentBox.y2 - sourceContentBox.y1);

            let menuHeight = menu.actor.get_allocation_box().y2 - menu.actor.get_allocation_box().y1;

            let x = sourceCenterX;
            let y = sourceCenterY;
            menu.x = x;
            menu.y = y - menuHeight;
        }));
        MyAppletMenu.addChildMenu(this.menu);
    },
    setActive : function () {
        // body...
    },
    /*change（自己加的）*/
    activate: function(event, keepMenu) {
        this.menu.toggle();
        this.emit('activate', event, keepMenu);
    }
};

function ShutdownButton(parent, hoverIcon) {
    this._init(parent, hoverIcon);
}

ShutdownButton.prototype = {
    _init: function(parent, hoverIcon) {
        this.actor = new St.BoxLayout({
            vertical: true
        });

        let box = new St.BoxLayout({
            style_class: 'menu-shutdown',
            vertical: false,
            reactive: true,
            track_hover: true
        });

        this.actor._delegate = this;
        this.parent = parent;
        this.hoverIcon = hoverIcon;

        this.shutdown = new St.Button({
            label: _('Shutdown')
        });
        this.shutdown.connect('clicked', Lang.bind(this, function(){
            Session.ShutdownRemote();
        }));
        this.shutdownMenu = new ShutdownMenuButton(this.parent, this.hoverIcon); 
        this.shutdownMenu.actor.set_style_class_name('menu-shutdown-button');
        box.add(this.shutdownMenu.actor);
        this.actor.add_actor(box);
    },

    setActive: function(active) {
        if (active) {
            //this.actor.remove_style_class_name('menu-shutdown-box');
            //this.actor.add_style_class_name('menu-shutdown-box-selected');

        } else {
            //this.actor.remove_style_class_name('menu-shutdown-box-selected');
            //this.actor.add_style_class_name('menu-shutdown-box');
        }
    }
};
/*
 * 便捷入口
 */
function ShortcutBox(menu, hoverIcon) {
    this._init(menu, hoverIcon);
}

ShortcutBox.prototype = {
    _init: function(menu, hoverIcon) {
        this.actor = new St.BoxLayout({
            style_class: 'right-buttons-box',
            vertical: true
        });

        this.actor._delegate = this;
        this.menu = menu; //和MyApplet的menu类型一致
        this.hoverIcon = hoverIcon;
        this.addItems();
    },

    addItems: function() {
        this.home = new ShortcutItem(_("Home"), "folder-home", "Util.spawnCommandLine('nemo')", this.menu, this.hoverIcon, false);
        this.documents = new ShortcutItem(_("Documents"), "folder-documents", "Util.spawnCommandLine('nemo "+_('Documents')+"')", this.menu, this.hoverIcon, false);
        this.pictures = new ShortcutItem(_("Pictures"), "folder-pictures", "Util.spawnCommandLine('nemo "+_('Pictures')+"')", this.menu, this.hoverIcon, false);
        this.music = new ShortcutItem(_("Music"), "folder-music", "Util.spawnCommandLine('nemo "+_('Music')+"')", this.menu, this.hoverIcon, false);
        this.videos = new ShortcutItem(_("Videos"), "folder-videos", "Util.spawnCommandLine('nemo "+_('Videos')+"')", this.menu, this.hoverIcon, false);
        this.computer = new ShortcutItem(_("Computer"), "computer", "Util.spawnCommandLine('nemo computer:///')", this.menu, this.hoverIcon, false);
        //this.packageItem = new Items.ShortcutItem(_("Package Manager"), "synaptic", "Util.spawnCommandLine('gksu synaptic')", this.menu, this.hoverIcon, false);
        this.control = new ShortcutItem(_("System Settings"), "control-center2", "Util.spawnCommandLine('cdos-settings')", this.menu, this.hoverIcon, false);
        //this.run = new ShortcutItem(_("Run"), "cs-xlet-running", "Main.getRunDialog().open()", this.menu, this.hoverIcon, false);
        this.terminal = new ShortcutItem(_("Terminal"), "terminal", "Util.spawnCommandLine('gnome-terminal')", this.menu, this.hoverIcon, false);
        this.help = new ShortcutItem(_("Help & Feedback"), "help", "Util.spawnCommandLine('cdoswelcome')", this.menu, this.hoverIcon, false);
        
        this.textBox = new St.BoxLayout({
            style_class: 'menu-text-box',
            vertical: true
        });
        this.actor.add_actor(this.textBox);

        this.textBox.add_actor(this.home.actor);
        this.textBox.add_actor(this.documents.actor);
        this.textBox.add_actor(this.pictures.actor);
        this.textBox.add_actor(this.music.actor);
        this.textBox.add_actor(this.videos.actor);
        let separator = new Separator();
        separator.actor.add_style_class_name('menu-separator-item');
        this.textBox.add_actor(separator.actor);

        this.textBox.add_actor(this.computer.actor);
        //this.textBox.add_actor(this.packageItem.actor);
        let separator = new Separator();
        separator.actor.add_style_class_name('menu-separator-item');
        this.textBox.add_actor(separator.actor);

        this.textBox.add_actor(this.control.actor);
        //this.textBox.add_actor(this.run.actor);
        this.textBox.add_actor(this.terminal.actor);
        this.textBox.add_actor(this.help.actor);       
    }
}

function FavoritesBox() {
    this._init();
}

FavoritesBox.prototype = {
    _init: function() {
        this.actor = new St.BoxLayout({
            vertical: true
        }); 

        this.actor._delegate = this;
        this._dragPlaceholder = null;
        this._dragPlaceholderPos = -1;
        this._animatingPlaceholdersCount = 0;
        this.actor.connect('style-changed', Lang.bind(this, this._change));
    },
    addActor: function(actor){
        this.actor.add_actor(actor);
    },
    _change: function () {
        let y = this.actor.get_allocation_box().y1;
        let y2;
        let parent = this.actor.get_parent();
        let children = this.actor.get_children();
        if(parent){
            y2 = parent.get_allocation_box().y2;
            for(let i = 0 ; i < children.length; i++){
                children[i].show();
                if((y + children[i].get_allocation_box().y2) > y2)
                    children[i].hide();
            }
        }
    },

    _clearDragPlaceholder: function() {
        if (this._dragPlaceholder) {
            this._dragPlaceholder.animateOutAndDestroy();
            this._dragPlaceholder = null;
            this._dragPlaceholderPos = -1;
        }
    },

    handleDragOver: function(source, actor, x, y, time) {
        let app = source.app;

        // Don't allow favoriting of transient apps
        if (app == null || app.is_window_backed() || (!(source instanceof FavoritesButton) && app.get_id() in AppFavorites.getAppFavorites().getFavoriteMap())) return DND.DragMotionResult.NO_DROP;

        let favorites = AppFavorites.getAppFavorites().getFavorites();
        let numFavorites = favorites.length;

        let favPos = favorites.indexOf(app);

        let children = this.actor.get_children();
        let numChildren = children.length;
        let boxHeight = this.actor.height;

        // Keep the placeholder out of the index calculation; assuming that
        // the remove target has the same size as "normal" items, we don't
        // need to do the same adjustment there.
        if (this._dragPlaceholder) {
            boxHeight -= this._dragPlaceholder.actor.height;
            numChildren--;
        }

        let pos = Math.round(y * numFavorites / boxHeight);

        if (pos != this._dragPlaceholderPos && pos <= numFavorites) {
            if (this._animatingPlaceholdersCount > 0) {
                let appChildren = children.filter(function(actor) {
                    return (actor._delegate instanceof FavoritesButton);
                });
                this._dragPlaceholderPos = children.indexOf(appChildren[pos]);
            } else {
                this._dragPlaceholderPos = pos;
            }

            // Don't allow positioning before or after self
            if (favPos != -1 && (pos == favPos || pos == favPos + 1)) {
                if (this._dragPlaceholder) {
                    this._dragPlaceholder.animateOutAndDestroy();
                    this._animatingPlaceholdersCount++;
                    this._dragPlaceholder.actor.connect('destroy', Lang.bind(this, function() {
                        this._animatingPlaceholdersCount--;
                    }));
                }
                this._dragPlaceholder = null;

                return DND.DragMotionResult.CONTINUE;
            }

            // If the placeholder already exists, we just move
            // it, but if we are adding it, expand its size in
            // an animation
            let fadeIn;
            if (this._dragPlaceholder) {
                this._dragPlaceholder.actor.destroy();
                fadeIn = false;
            } else {
                fadeIn = true;
            }

            this._dragPlaceholder = new DND.GenericDragPlaceholderItem();
            this._dragPlaceholder.child.set_width(source.actor.height);
            this._dragPlaceholder.child.set_height(source.actor.height);
            this.actor.insert_actor(this._dragPlaceholder.actor, this._dragPlaceholderPos);
            if (fadeIn) this._dragPlaceholder.animateIn();
        }

        let srcIsFavorite = (favPos != -1);

        if (srcIsFavorite) return DND.DragMotionResult.MOVE_DROP;

        return DND.DragMotionResult.COPY_DROP;
    },

    // Draggable target interface
    acceptDrop: function(source, actor, x, y, time) {
        let app = source.app;

        // Don't allow favoriting of transient apps
        if (app == null || app.is_window_backed()) {
            return false;
        }

        let id = app.get_id();

        let favorites = AppFavorites.getAppFavorites().getFavoriteMap();

        let srcIsFavorite = (id in favorites);

        let favPos = 0;
        let children = this.actor.get_children();
        for (let i = 0; i < this._dragPlaceholderPos; i++) {
            if (this._dragPlaceholder && children[i] == this._dragPlaceholder.actor) continue;

            if (!(children[i]._delegate instanceof FavoritesButton)) continue;

            let childId = children[i]._delegate.app.get_id();
            if (childId == id) continue;
            if (childId in favorites) favPos++;
        }

        Meta.later_add(Meta.LaterType.BEFORE_REDRAW, Lang.bind(this, function() {
            let appFavorites = AppFavorites.getAppFavorites();
            if (srcIsFavorite) appFavorites.moveFavoriteToPos(id, favPos);
            else appFavorites.addFavoriteAtPos(id, favPos);
            return false;
        }));

        return true;
    }
}
Signals.addSignalMethods(FavoritesBox.prototype);

function Separator(params) {
    this._init(params);
}

Separator.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function(params) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            reactive: false
        });
        let style_class = params && params['style_class'] ? params.style_class :  'menu-separator-line';
        this.actor.style = 'padding: 0;';
        this._drawingArea = new St.DrawingArea({
            style_class: style_class
        });
        this.addActor(this._drawingArea, {
            span: -1,
            expand: true
        });
        this._drawingArea.connect('repaint', Lang.bind(this, this._onRepaint));
    },

    _onRepaint: function(area) {
        let cr = area.get_context();
        let themeNode = area.get_theme_node();
        let [width, height] = area.get_surface_size();
        let margin = themeNode.get_length('-margin-horizontal');
        let gradientHeight = themeNode.get_length('-gradient-height');
        let startColor = themeNode.get_color('-gradient-start');
        let endColor = themeNode.get_color('-gradient-end');

        let gradientWidth = (width - margin * 2);
        let gradientOffset = (height - gradientHeight) / 2;
        let pattern = new Cairo.LinearGradient(margin, gradientOffset, width - margin, gradientOffset + gradientHeight);
        pattern.addColorStopRGBA(0, startColor.red / 255, startColor.green / 255, startColor.blue / 255, startColor.alpha / 255);
        pattern.addColorStopRGBA(0.5, endColor.red / 255, endColor.green / 255, endColor.blue / 255, endColor.alpha / 255);
        pattern.addColorStopRGBA(1, startColor.red / 255, startColor.green / 255, startColor.blue / 255, startColor.alpha / 255);
        cr.setSource(pattern);
        cr.rectangle(margin, gradientOffset, gradientWidth, gradientHeight);
        cr.fill();
    }
};

/*
 * Applet.AppletPopupMenu
*/
function AppletPopupMenu(launcher, orientation) {
    this._init(launcher, orientation);
}

AppletPopupMenu.prototype = {
    __proto__: PopupMenu.PopupMenu.prototype,

    _init: function(launcher, orientation) {
        PopupMenu.PopupMenu.prototype._init.call(this, launcher.actor, 0.0, orientation, 0);
        this._boxPointer._reposition = function(sourceActor, alignment) {
            // Position correctly relative to the sourceActor
            let sourceNode = sourceActor.get_theme_node();
            let sourceContentBox = sourceNode.get_content_box(sourceActor.get_allocation_box());
            let sourceAllocation = Cinnamon.util_get_transformed_allocation(sourceActor);
            let sourceCenterX = sourceAllocation.x1 + sourceContentBox.x1 + (sourceContentBox.x2 - sourceContentBox.x1) * this._sourceAlignment;
            let sourceCenterY = sourceAllocation.y1 + sourceContentBox.y1 + (sourceContentBox.y2 - sourceContentBox.y1) * this._sourceAlignment;
            let [minWidth, minHeight, natWidth, natHeight] = this.actor.get_preferred_size();

            // We also want to keep it onscreen, and separated from the
            // edge by the same distance as the main part of the box is
            // separated from its sourceActor
            let monitor = Main.layoutManager.findMonitorForActor(sourceActor);
            let themeNode = this.actor.get_theme_node();
            let borderWidth = themeNode.get_length('-arrow-border-width');
            let arrowBase = themeNode.get_length('-arrow-base');
            let borderRadius = themeNode.get_length('-arrow-border-radius');
            let margin = (4 * borderRadius + borderWidth + arrowBase);
            let halfMargin = margin / 2;

            let themeNode = this.actor.get_theme_node();
            let gap = themeNode.get_length('-boxpointer-gap');

            let resX, resY;

            switch (this._arrowSide) {
                case St.Side.TOP:
                    resY = sourceAllocation.y2 + gap;
                    break;
                case St.Side.BOTTOM:
                    resY = sourceAllocation.y1 - natHeight - gap;
                    break;
                case St.Side.LEFT:
                    resX = sourceAllocation.x2 + gap;
                    break;
                case St.Side.RIGHT:
                    resX = sourceAllocation.x1 - natWidth - gap;
                    break;
            }

            // Now align and position the pointing axis, making sure
            // it fits on screen
            switch (this._arrowSide) {
                case St.Side.TOP:
                case St.Side.BOTTOM:
                    resX = sourceCenterX - (halfMargin + (natWidth - margin) * alignment);
                    resX = Math.max(resX, monitor.x + 10);
                    resX = Math.min(resX, monitor.x + monitor.width - (10 + natWidth));
                    this.setArrowOrigin(sourceCenterX - resX);
                    break;

                case St.Side.LEFT:
                case St.Side.RIGHT:
                    resY = sourceCenterY - (halfMargin + (natHeight - margin) * alignment);

                    resY = Math.max(resY, monitor.y + 10);
                    resY = Math.min(resY, monitor.y + monitor.height - (10 + natHeight));

                    this.setArrowOrigin(sourceCenterY - resY);
                    break;
            }

            let parent = this.actor.get_parent();
            let success, x, y;
            while (!success) {
                [success, x, y] = parent.transform_stage_point(resX, resY);
                parent = parent.get_parent();
            }
            if( typeof this.x_axis == 'number') x = this.x_axis;
            if( typeof this.y_axis == 'number') y = this.y_axis;
            this._xPosition = Math.floor(x);
            this._yPosition = Math.floor(y);
            this._shiftActor();
        }
        this._boxPointer.setPosition = function(sourceActor, alignment, x, y) {
            // We need to show it now to force an allocation,
            // so that we can query the correct size.
            this.x_axis = x;
            this.y_axis = y;

            this.actor.show();

            this._sourceActor = sourceActor;
            this._arrowAlignment = alignment;

            this._reposition(sourceActor, alignment);
        },
        Main.uiGroup.add_actor(this.actor);
        this.box.style = 'padding: 0px';
        this.actor.hide();
    },
    open: function(animate) {
        if (this.isOpen)
            return;

        this.emit('before-open', true);
        this.setMaxHeight();

        this.isOpen = true;

        if (global.menuStackLength == undefined)
            global.menuStackLength = 0;
        global.menuStackLength += 1;
        this._boxPointer.setPosition(this.sourceActor, this._arrowAlignment, this.x, this.y);
        this._boxPointer.show(animate);

        this.actor.raise_top();

        this.emit('open-state-changed', true);
    },
    toggle: function() {
        if (this.isOpen) {
            this.close();
        } else {
            this.open();
        }
    },
    
    set xOffset(offset) {
        this._boxPointer._xOffset = offset;
        this._boxPointer._shiftActor();
    },

    get xOffset() {
        return this._boxPointer._xOffset;
    },

    set yOffset(offset) {
        this._boxPointer._yOffset = offset;
        this._boxPointer._shiftActor();
    },

    get yOffset() {
        return this._boxPointer._yOffset;
    },

    set position(position) {
        this._boxPointer._xPosition = Math.floor(position[0]);
        this._boxPointer._yPosition = Math.floor(position[1]);
        this._boxPointer._shiftActor();
    }
}

/*
 * menu 右键菜单类
 */

function Contextmenu(launcher) {
    this._init(launcher);
}

Contextmenu.prototype = {
    __proto__: AppletPopupMenu.prototype,
    _init: function(launcher) {
        let orientation = 3;
        this.launcher = launcher;
        AppletPopupMenu.prototype._init.call(this, this.launcher, orientation);
        this.connect('open-state-changed', Lang.bind(this, function (menu, isOpen) {
            if(!isOpen){
                let menu = this.launcher.appsMenuButton.menu;
                Mainloop.timeout_add(300, Lang.bind(this, function () {
                    global.stage.set_key_focus(menu.actor);
                }));
                
            }else{
                //log(global.stage.get_key_focus());
            }
        }));
    }
}

/*
* SearchResultButton
*/
function SearchResultButton(appsMenuButton, app) {
    this._init(appsMenuButton, app);
}

SearchResultButton.prototype = {
    __proto__: GenericApplicationButton.prototype,

    _init: function(appsMenuButton, app) {
        GenericApplicationButton.prototype._init.call(this, appsMenuButton, app, true);
        this.actor.add_style_class_name('menu-searchs-button');
        if (app) {
            this.icon = this.app.create_icon_texture(APPLICATION_ICON_SIZE);
            this.label = new St.Label({
                text: this.app.get_name(),
                style_class: 'menu-application-button-label'
            });
            this.addActor(this.icon);
            this.addActor(this.label);
        }
        this.actor.connect('leave-event', Lang.bind(this, function() {
            this.actor.remove_style_class_name("menu-searchs-button-selected");
        }));
        this.actor.connect('enter-event', Lang.bind(this, function() {
            this.actor.add_style_class_name("menu-searchs-button-selected");
        }));
    },

    get_app_id: function() {
        return this.app.get_id();
    }
};

/*****************************  MyApplet  **************************/
function MyApplet(orientation, panel_height, instance_id) {
    this._init(orientation, panel_height, instance_id);
}


MyApplet.prototype = {
    __proto__: Applet.TextIconApplet.prototype,

    _init: function(orientation, panel_height, instance_id) {
        Applet.TextIconApplet.prototype._init.call(this, orientation, panel_height, instance_id);

        try {
	    this.actor.remove_style_class_name("applet-box");
	    this._applet_icon_box.set_margin_left(8);
            this.settings = new Settings.AppletSettings(this, "menu@cdos.org", instance_id);
            this.set_applet_tooltip(_("Start"));

            ori = orientation; //change（自己加的）

            menuManager = new PopupMenu.PopupMenuManager(this); //change

            //this.menu = new Applet.AppletPopupMenu(this, orientation);  //创建菜单
            this.menu = new AppletPopupMenu(this, orientation);

            MyAppletMenu = this.menu; //change（自己加的）

            this.menu.connect('before-open', function (menu) {
                menu.x  = 0;
            });

            this.actor.add_style_class_name('menu'); //change（自己加的），开始菜单按钮         ///////调试       
            //this.menuManager.addMenu(this.menu);   //在菜单管理器中添加菜单
            menuManager.addMenu(this.menu); //change

            this.actor.connect('key-press-event', Lang.bind(this, this._onSourceKeyPress));

            this.settings.bindProperty(Settings.BindingDirection.IN, "show-recent", "showRecent", this._refreshRecent, null); //change（自己加的）

            this.menu._arrowSide = 0; //change（自己加的）    useless
            this.menu.setColumnWidths(0); //change（自己加的）  useless
            this.menu.actor.set_opacity(0.1); //change（自己加的） useless
            this.menu.box.add_style_class_name('menu-background'); //change

            this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateChanged));
            this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateToggled));

            this._updateIcon();

            global.settings.connect("changed::menu-icon", Lang.bind(this, function() {
                this._updateIcon();
            }));

            this._searchInactiveIcon = new St.Icon({
                style_class: 'menu-search-entry-icon',
                icon_name: 'edit-find',
                icon_type: St.IconType.SYMBOLIC
            });
            this._searchActiveIcon = new St.Icon({
                style_class: 'menu-search-entry-icon',
                icon_name: 'edit-clear',
                icon_type: St.IconType.SYMBOLIC
            });
            this._searchTimeoutId = 0;
            this._searchIconClickedId = 0;
            this._applicationsButtons = new Array();
            this._favoritesButtons = new Array();
            this._selectedItemIndex = null;
            this._previousSelectedItemIndex = null;
            this._activeContainer = null;
            this._recentButtons = new Array();
            this._locked = false;

            this.RecentManager = new DocInfo.DocManager(); //change（自己加的）
            this.RecentManager.connect('changed', Lang.bind(this, this._refreshRecent)); //change（自己加的）
            this._display();
            appsys.connect('installed-changed', Lang.bind(this, this._refreshApps));
            AppFavorites.getAppFavorites().connect('changed', Lang.bind(this, function (appfavs, oldVal, newVal) {
                this._locked = true;
                this._refreshFavs();
                this._refreshRecent();
            }));

            this.hover_delay = global.settings.get_int("menu-hover-delay") / 1000;

            global.settings.connect("changed::menu-hover-delay", Lang.bind(this, function() {
                this.hover_delay = global.settings.get_int("menu-hover-delay") / 1000;
            }));
            
            global.display.connect('overlay-key', Lang.bind(this, function () {
                try {
                    this.menu.toggle();
                } catch (e) {
                    global.logError(e);
                }
            }));
            
            this.edit_menu_item = new Applet.MenuItem(_("Edit menu"), Gtk.STOCK_EDIT, Lang.bind(this, this._launch_editor));
            this._applet_context_menu.addMenuItem(this.edit_menu_item);

        } catch (e) {
            global.logError(e);
        }
    },

    on_orientation_changed: function(orientation) { //改变方向
        this.menu.destroy();
        this.menu = new Applet.AppletPopupMenu(this, orientation);
        menuManager.addMenu(this.menu); //change

        this.menu.actor.add_style_class_name('menu-background');
        this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateChanged));
        this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateToggled));

        this._display();
    },

    _launch_editor: function() {
        Util.spawnCommandLine("cdos-menu-editor");
    },

    on_applet_clicked: function(event) {
        this.menu.toggle();
    },

    _onSourceKeyPress: function(actor, event) {
        let symbol = event.get_key_symbol();

        if (symbol == Clutter.KEY_space || symbol == Clutter.KEY_Return) { //按下空格或回车
            this.menu.toggle(); //切换菜单
            return true;
        } else if (symbol == Clutter.KEY_Escape && this.menu.isOpen) { //按下Esc
            this.menu.close(); //关闭菜单
            return true;
        } else if (symbol == Clutter.KEY_Down) { //按下下键
            if (!this.menu.isOpen) this.menu.toggle();
            this.menu.actor.navigate_focus(this.actor, Gtk.DirectionType.DOWN, false);
            return true;
        } else return false;
    },

    _onOpenStateChanged: function(menu, open) {
        if (open) {
            this.actor.add_style_pseudo_class('active');
            this._locked = false;
        } else {
            this.actor.remove_style_pseudo_class('active');
            if (this.searchActive) {
                this.resetSearch();
            }
        }
    },

    _onOpenStateToggled: function(menu, open) {
        if (open) {
            global.stage.set_key_focus(this.searchEntry);
            this._selectedItemIndex = null;
            this._activeContainer = null;
            this._appletStyles();
           //this.reset();
        } else {
            this.closeApplicationsContextMenus(null, false);
            Mainloop.timeout_add(200, Lang.bind(this, this.reset));
            //this.resetSearch();
        }
    },

    destroy: function() {
        this.actor._delegate = null;
        this.menu.destroy();
        this.actor.destroy();
        this.emit('destroy');
    },

    _updateIcon: function() {
        let icon_file = global.settings.get_string("menu-icon");
        if (!icon_file) {
            icon_file = global.datadir + '/theme/menu.svg';
        }
        if (!GLib.file_test(icon_file, GLib.FileTest.EXISTS))
            icon_file = global.datadir + '/theme/menu.png';
        try {
            this.set_applet_icon_path(icon_file);
        } catch (e) {
            global.log("WARNING : Could not load icon file \"" + icon_file + "\" for menu button");
        }
    },

    _onMenuKeyPress: function(actor, event) {

        let symbol = event.get_key_symbol();

        if (symbol == Clutter.KEY_Super_L && this.menu.isOpen) { //开始菜单在打开状态下按super键关闭
            this.menu.close();
            return true;
        }

        if (this._activeContainer === null && symbol == Clutter.KEY_Up) {
            this._activeContainer = this.categoriesBox;
            children = this._activeContainer.get_children();
            this._selectedItemIndex = children.length;
        } else if (this._activeContainer === null && symbol == Clutter.KEY_Down) {
            this._activeContainer = this.categoriesBox;
            children = this._activeContainer.get_children();
            this._selectedItemIndex = -1;
        } else if (this._activeContainer === null) {
            this._activeContainer = this.categoriesBox;
            this._selectedItemIndex = -1;
            this._previousSelectedItemIndex = -1;
        }


        let children = this._activeContainer.get_children();

        if (children.length == 0) {
            this._activeContainer = this.categoriesBox;
            this._selectedItemIndex = -1;
            this._previousSelectedItemIndex = -1;
            children = this._activeContainer.get_children();
        }

        let index = this._selectedItemIndex;

        if (symbol == Clutter.KEY_Up) {
            if (this._activeContainer == this.categoriesBox) index = this._selectedItemIndex - 1 < 0 ? 0 : this._selectedItemIndex - 2;
            else index = this._selectedItemIndex - 1 < 0 ? 0 : this._selectedItemIndex - 1;
        } else if (symbol == Clutter.KEY_Down) {
            if (this._activeContainer == this.categoriesBox && this._selectedItemIndex != -1) index = this._selectedItemIndex + 2 >= children.length ? children.length - 2 : this._selectedItemIndex + 2;
            else index = this._selectedItemIndex + 1 == children.length ? children.length - 1 : this._selectedItemIndex + 1;
        } else if (symbol == Clutter.KEY_Right && this._activeContainer === this.categoriesBox) {
            this._activeContainer = this.categoriesBox;
            children = this._activeContainer.get_children();
            index = 0;
            this._previousSelectedItemIndex = this._selectedItemIndex;
            this._selectedItemIndex = -1;
        } else {
            return false;
        }

        if (index == this._selectedItemIndex) {
            return true;
        }

        if (this._activeContainer == this.categoriesBox) {
            if (index >= children.length - 1) index = children.length - 2;
        } else {
            if (index >= children.length) index = children.length - 1;
        }

        this._selectedItemIndex = index;
        let item_actor = children[this._selectedItemIndex];

        if (!item_actor || item_actor === this.searchEntry) {
            return false;
        }

        item_actor._delegate.emit('enter-event');
        return true;
    },

    _addEnterEvent: function(button, callback) { //鼠标进入事件
        let _callback = Lang.bind(this, function() {
            let parent = button.actor.get_parent(); //获取按钮的父菜单
            if (this._activeContainer === this.categoriesBox && parent !== this._activeContainer) {
                this._previousSelectedItemIndex = this._selectedItemIndex;
            }
            this._activeContainer = parent;
            let children = this._activeContainer.get_children();
            for (let i = 0, l = children.length; i < l; i++) {
                if (button.actor === children[i]) {
                    this._selectedItemIndex = i;
                }
            };
            callback();
        });
        button.connect('enter-event', _callback);
        button.actor.connect('enter-event', _callback);
    },

    _clearSelections: function(container) {
        container.get_children().forEach(function(actor) {
            actor.style_class = "menu-category-button";
            if (actor._delegate && actor._delegate.menu)
                actor._delegate.menu.close(true);
        });
    },

    _appletStyles: function() {
        if(this._locked) return;
        // 20151210
        let recents = this.RecentManager._infosByTimestamp.length;
        if(recents > MAX_RECENT_FILES) recents = MAX_RECENT_FILES;

        // favorites 
        let launchers = global.settings.get_strv('favorite-apps');
        //设置小程序的尺寸
        let monitorHeight = Main.layoutManager.primaryMonitor.height;
        //let height = Math.floor(monitorHeight * 52 / 100);
        let panelHeight = 50;
        let width = 424;
        let height = 522;
        
        let leftwidth = 262;
        let padding = 4;
        let space = 8;
        let nums = (launchers.length + recents);
        let maxHeight = monitorHeight - panelHeight - space * 2 - this.bottom.height - 18;

        if( nums > 9){
            height += (nums - 9) * 44;
        }
        if(height > maxHeight)
            height = maxHeight;

        this.top.width = width;
        this.leftPaneBox.width = leftwidth;
        this.rightPaneBox.width = width-leftwidth;
        this.bottomLeft.width = leftwidth;
        this.bottomRight.width = width-leftwidth;
        this.favBoxWith = (this.leftPaneBox.get_allocation_box().x2 - this.leftPaneBox.get_allocation_box().x1); //原代码      
        let scrollBoxHeight = -padding + height - (this.appsButtonBox.get_allocation_box().y2 - this.appsButtonBox.get_allocation_box().y1);

        this.categoriesScrollBox.set_style("width: " + this.favBoxWith + "px; height: " + scrollBoxHeight + "px;");
        this.favsBox.set_style("height: "+ scrollBoxHeight +"px; overflow:hidden;");

        let topHeight = (this.leftPaneBox.get_allocation_box().y2 - this.leftPaneBox.get_allocation_box().y1);
        this.searchResultBox.set_style("height: " + topHeight + "px;");
        return;
        //设置菜单右侧部分的高度
        let rightPaneBoxHeight = this.rightPaneBox.get_allocation_box().y2 - this.rightPaneBox.get_allocation_box().y1;
        let hoverIconHeight = this.hoverIcon.actor.get_allocation_box().y2 - this.hoverIcon.actor.get_allocation_box().y1;
        let shutdownBoxHeight = this.shutdownBox.actor.get_allocation_box().y2 - this.shutdownBox.actor.get_allocation_box().y1;

        let shortcutBoxHeight = rightPaneBoxHeight - hoverIconHeight - shutdownBoxHeight - 12;
        this.ShortcutBox.actor.style = 'height:' + shortcutBoxHeight + 'px';
    },

    _refreshApps: function() {
        this._applicationsButtons = new Array();

        //Remove all categories
        this.categoriesBox.get_children().forEach(Lang.bind(this, function(child) {
            child.destroy();
        }));
        let favoritesBox = new FavoritesBox();
        this._allAppsLabel = new St.Label({
            style_class: 'favorites-label',
            text: _('All Applications')
        });
        let separator = new Separator({
            style_class : 'menu-favorites-separator'
        });
        favoritesBox.addActor(this._allAppsLabel);
        favoritesBox.addActor(separator.actor);
        favoritesBox.actor.set_style('height:38px; padding:0;');
        this.categoriesBox.add_actor(favoritesBox.actor);

        //let trees = [appsys.get_tree(), appsys.get_settings_tree()];
        let trees = [appsys.get_tree()];
        for (let i in trees) {
            let tree = trees[i];
            let root = tree.get_root_directory();

            let iter = root.iter();
            let nextType;
            while ((nextType = iter.next()) != CMenu.TreeItemType.INVALID) {
                if (nextType == CMenu.TreeItemType.DIRECTORY) {
                    let dir = iter.get_directory();
                    if (dir.get_is_nodisplay() || dir.get_menu_id() === 'Preferences') continue;
                    this.applicationsByCategory[dir.get_menu_id()] = new Array();
                    this.categoryMap[dir.get_menu_id()] = dir.get_name();
                    this._loadCategory(dir);
                    if (this.applicationsByCategory[dir.get_menu_id()].length > 0) {
                        let categoryButton = new CategoryButton(dir);
                        categoryButton.actor.connect('enter-event', function() {
                            categoryButton.isHovered = true;
                            categoryButton.actor.add_style_class_name("menu-category-button-selected");
                        });
                        categoryButton.actor.connect('leave-event', function() {
                            categoryButton.isHovered = false;
                            if (!categoryButton.menu.isOpen)
                                categoryButton.actor.remove_style_class_name("menu-category-button-selected");
                        });
                        this.categoriesBox.add_actor(categoryButton.actor);
                        this.categoriesBox.add_actor(categoryButton.menu.actor);
                        this._displayButtons(this._listApplications(dir.get_menu_id()), null, categoryButton, true);
                    }
                }
            }
        }
    },

    _refreshFavs: function() {
        global.stage.set_key_focus(this.menu.actor);
        //Remove all favorites
        let i = 0;
        this.favsBox.get_children().forEach(Lang.bind(this, function(child) {
            if (i == 0) {
                child.destroy();
                return;
            }
        }));

        let favoritesBox = new FavoritesBox();
        this.favsBox.add_actor(favoritesBox.actor, {
            y_align: St.Align.END,
            y_fill: false
        });

        /*change（自己加的）*/
        this.favsLabel = new St.Label({
            style_class: 'favorites-label',
            text: _('Quick Operation')
        });
        favoritesBox.addActor(this.favsLabel); //添加快捷操作标签
        let separator = new Separator({
            style_class : 'menu-favorites-separator'
        });
        favoritesBox.addActor(separator.actor);
        //Load favorites again
        this._favoritesButtons = new Array();
        //20151210
        let recents = this.RecentManager._infosByTimestamp.length;
        if(recents > MAX_RECENT_FILES) recents = MAX_RECENT_FILES;

        let launchers = global.settings.get_strv('favorite-apps');
        let appSys = Cinnamon.AppSystem.get_default();
        let j = 0;

        for (let i = 0; i < launchers.length; i++) {
            let app = appSys.lookup_app(launchers[i]);
            if (!app) app = appSys.lookup_settings_app(launchers[i]);
            if (app) {
                let button = new FavoritesButton(this, app, launchers.length); // + 3 because we're adding 3 system buttons at the bottom
                this._favoritesButtons[app] = button;
                favoritesBox.addActor(button.actor);
                button.actor.connect('enter-event', Lang.bind(this, function() {
                    //
                }));
                button.actor.connect('leave-event', Lang.bind(this, function() {
                    //
                }));
                ++j;
            }
        }
    },
    /*change(自己加的)*/
    /*最近常用*/
    _refreshRecent: function() {
        let i = 0;
        this.favsBox.get_children().forEach(Lang.bind(this, function(child) {
            if (i > 0) child.destroy();
            i++;
        }));

        let favoritesBox = new FavoritesBox();
        this.favsBox.add_actor(favoritesBox.actor);

        /*change（自己加的）*/
        this.recentUsedLabel = new St.Label({
            style_class: 'favorites-label',
            text: _('Recent Used')
        });
        favoritesBox.addActor(this.recentUsedLabel); //添加最近常用标签
        let separator = new Separator({
            style_class : 'menu-favorites-separator'
        });
        favoritesBox.addActor(separator.actor);

        for (let i = 0; i < this._recentButtons.length; i++) {
            this._recentButtons[i].actor.destroy(); //清除recentButtons
        }
        this._recentButtons = new Array();
        if (this.RecentManager._infosByTimestamp.length > 0) {
            for (let id = 0; id < MAX_RECENT_FILES && id < this.RecentManager._infosByTimestamp.length; id++) {
                let button = new RecentButton(this, this.RecentManager._infosByTimestamp[id]);
                this._addEnterEvent(button, Lang.bind(this, function() { //鼠标进入事件
                    //
                }));
                button.actor.connect('leave-event', Lang.bind(this, function() { //鼠标离开事件
                    //
                }));
                this._recentButtons.push(button);
                favoritesBox.addActor(button.actor);
            }
        }
    },
    _loadCategory: function(dir, top_dir) {
        var iter = dir.iter();
        var nextType;
        if (!top_dir) top_dir = dir;
        while ((nextType = iter.next()) != CMenu.TreeItemType.INVALID) {
            if (nextType == CMenu.TreeItemType.ENTRY) {
                var entry = iter.get_entry();
                if (!entry.get_app_info().get_nodisplay()) {
                    var app = appsys.lookup_app_by_tree_entry(entry);
                    if (!app) app = appsys.lookup_settings_app_by_tree_entry(entry);
                    if (!this.applicationsByCategory[top_dir.get_menu_id()]){
                        this.applicationsByCategory[top_dir.get_menu_id()] = new Array();
                        this.categoryMap[top_dir.get_menu_id] = top_dir.get_name();
                    }
                    this.applicationsByCategory[top_dir.get_menu_id()].push(app);
                }
            } else if (nextType == CMenu.TreeItemType.DIRECTORY) {
                this._loadCategory(iter.get_directory(), top_dir);
            }
        }
    },

    _display: function() {
        this._activeContainer = null;
        let section = new PopupMenu.PopupMenuSection();
        this.menu.addMenuItem(section);

        this.mainBox = new St.BoxLayout({
            style_class: 'menu-applications-box',
            vertical: true
        });

        /**top**/
        this.top = new St.Bin({x_fill: true});
        //defaultView
        this.defaultView = new St.BoxLayout({
            vertical: false
        });
        //
        this.leftPaneBox = new St.BoxLayout({
            style_class: 'menu-left-pane',
            vertical: true
        });
        // favorites
        this.favsBox = new St.BoxLayout({
            vertical: true,
            style_class: 'menu-favorites-box'
        });

        //this.leftPane = new St.Bin({x_fill: true});
        this.leftPane = new  St.BoxLayout({vertical: true});
        //categories
        this.categoriesBox = new St.BoxLayout({
            style_class: 'menu-categories-box',
            vertical: true
        });
        this.categoriesScrollBox = new St.ScrollView({
            x_fill: true,
            y_fill: false,
            y_align: St.Align.START,
            style_class: 'vfade menu-applications-scrollbox'
        });

        let vscroll = this.categoriesScrollBox.get_vscroll_bar();
        vscroll.connect('scroll-start', Lang.bind(this, function() {
            this.menu.passEvents = true;
        }));
        vscroll.connect('scroll-stop', Lang.bind(this, function() {
            this.menu.passEvents = false;
        }));
        this.categoriesScrollBox.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
        this.categoriesScrollBox.add_actor(this.categoriesBox);
        this.categoriesScrollBox.hide();
        //默认显示快捷操作
        this.leftPane.add_actor(this.favsBox);
        this.leftPane.add_actor(this.categoriesScrollBox);
        //所有程序按钮
        this.allProButton = new AllProgramsButton(_("All Programs"), this, false);
        this.appsButtonBox = new St.BoxLayout({
            style_class: 'menu-appps-button-box',
            vertical: true
        });
        this.appsButtonBox.add_actor(this.allProButton.actor);

        this.leftPaneBox.add_actor(this.leftPane); //
        this.leftPaneBox.add_actor(new Separator().actor); //
        this.leftPaneBox.add_actor(this.appsButtonBox); //
        //右边栏
        this.rightPaneBox = new St.BoxLayout({
            style_class: 'menu-right-pane',
            vertical: true
        });
        //头像
        this.hoverIcon = new HoverIcon(this.menu);
        //快捷入口
        this.ShortcutBox = new ShortcutBox(this.menu, this.hoverIcon);
        this.rightPaneBox.add_actor(this.hoverIcon.actor);
        this.rightPaneBox.add_actor(this.ShortcutBox.actor);

        this.defaultView.add_actor(this.leftPaneBox);
        this.defaultView.add_actor(this.rightPaneBox);
        /*searchView*/
        this.searchView = new St.BoxLayout({
            vertical: true
        });
        this.searchResultBox = new St.BoxLayout({
            style_class: 'menu-search-result-box',
            vertical: true
        });
        this.searchView.add_actor(this.searchResultBox);
        //
        this.top.set_child(this.defaultView, {
            x_align: St.Align.START,
            x_fill: true
        });
        /*bottom*/
        this.bottom = new St.BoxLayout({
            vertical: false
        });
        this.bottomLeft = new St.BoxLayout({
            vertical: true
        });
        //search
        this.searchBox = new St.BoxLayout({ //searchBox
            style_class: 'menu-search-box',
            vertical: true
        });
        this.searchEntry = new St.Entry({ //searchEntry
            name: 'menu-search-entry',
            style_class: 'menu-search-entry',
            hint_text: _("Type to search..."),
            track_hover: true,
            can_focus: true
        });
        this.searchBox.add_actor(this.searchEntry);
        this.searchEntry.set_secondary_icon(this._searchInactiveIcon);
        this.searchActive = false;
        this.searchEntryText = this.searchEntry.clutter_text;
        this.searchEntryText.connect('text-changed', Lang.bind(this, this._onSearchTextChanged));
        this.searchEntryText.connect('key-press-event', Lang.bind(this, this._onMenuKeyPress));
        this._previousSearchPattern = "";
        this.bottomLeft.add_actor(this.searchBox);

        this.bottomRight = new St.BoxLayout({
            vertical: true
        });
        // shutdown
        this.shutdownBox = new St.BoxLayout({
            style_class: 'menu-shutdown-box',
            vertical: false
        });
        this.shutdownButton = new ShutdownButton(this, this.hoverIcon);
        this.shutdownBox.add_actor(this.shutdownButton.actor);
        this.bottomRight.add_actor(this.shutdownBox);

        this.bottom.add_actor(this.bottomLeft);
        this.bottom.add_actor(this.bottomRight);

        this.mainBox.add_actor(this.top, {
            y_align: St.Align.START,
            y_fill: false
        });
        this.mainBox.add_actor(this.bottom, {
            y_align: St.Align.END,
            y_fill: false
        });
        section.actor.add_actor(this.mainBox);

        this.applicationsByCategory = {};
        this.categoryMap = {};
        this._refreshFavs();
        this._refreshRecent();
        this._refreshApps();

        this.selectedAppBox = new St.BoxLayout({
            style_class: 'menu-selected-app-box',
            vertical: true
        });
        this.selectedAppBox.hide();
        this.selectedAppTitle = new St.Label({
            style_class: 'menu-selected-app-title',
            text: ""
        });
        this.selectedAppDescription = new St.Label({
            style_class: 'menu-selected-app-description',
            text: ""
        });
        this.selectedAppBox.add_actor(this.selectedAppTitle);
        this.selectedAppBox.add_actor(this.selectedAppDescription);
    },

    _clearApplicationsBox: function(selectedActor) {
        let actors = this.categoriesBox.get_children();
        for (let i = 0; i < actors.length; i++) {
            let actor = actors[i];
            if (actor._delegate && actor._delegate.menu){
                actor._delegate.removeItems();
                actor._delegate.menu.closex(); 
            }
        }
    },

    _select_category: function(dir, categoryButton) {
        this.resetSearch();
        //this._clearApplicationsBox(categoryButton.actor);
        if (dir)
            this._displayButtons(this._listApplications(dir.get_menu_id()), null, categoryButton, true);
        else 
            this._displayButtons(this._listApplications(null), null, categoryButton);
        this.closeApplicationsContextMenus(null, false);
    },

    closeApplicationsContextMenus: function(excludeApp, animate) {
        for (var app in this._applicationsButtons) {
            if (app != excludeApp && this._applicationsButtons[app].menu.isOpen) {
                if (animate) this._applicationsButtons[app].toggleMenu();
                else this._applicationsButtons[app].closeMenu();
            }
        }

        for (var fav in this._favoritesButtons) {
            if (fav != excludeApp && this._favoritesButtons[fav].menu.isOpen) {
                if (animate) this._favoritesButtons[fav].toggleMenu();
                else this._favoritesButtons[fav].closeMenu();
            }
        }
    },

    _displayButtons: function(apps, places, categoryButton, isSub) {
        if (apps) {
            categoryButton.menu.actor.height = 2;
            for (var i = 0; i < apps.length; i++) {
                let app = apps[i];
                if (!this._applicationsButtons[app]) {
                    let applicationButton = new ApplicationButton(this, app);

                    if(isSub)
                        applicationButton.actor.add_style_class_name('sub');

                    applicationButton.actor.connect('leave-event', Lang.bind(this, function() {
                        applicationButton.actor.remove_style_class_name("menu-application-button-selected");
                    }));
                    applicationButton.actor.connect('enter-event', Lang.bind(this, function() {
                        applicationButton.actor.add_style_class_name("menu-application-button-selected");
                    }));

                    this._applicationsButtons[app] = applicationButton;
                }
                categoryButton.menu.addMenuItem(this._applicationsButtons[app]);
                categoryButton.menu.actor.height += 30;
            }
        }

        if (places) {
            for (var i = 0; i < places.length; i++) {
                let place = places[i];
                let button = new PlaceButton(this, place, place.name);
                this._addEnterEvent(button, Lang.bind(this, function() {
                    //this._clearSelections(this.categoriesBox);
                    button.actor.add_style_class_name("menu-category-button-selected");
                }));
                button.actor.connect('leave-event', Lang.bind(this, function() {
                    button.actor.remove_style_class_name("menu-category-button-selected");
                }));
                categoryButton.menu.box.add_actor(button.actor);
            }
        }
    },

    _select_places: function(button) {
        this.resetSearch();
        this._clearApplicationsBox(button.actor);

        let bookmarks = this._listBookmarks();
        let devices = this._listDevices();
        this._displayButtons(null, bookmarks.concat(devices), button);
    },
/*
 * 
*/
    _categoryButtonsHidden: function(hide) {
        let categoriesBoxChilren = this.categoriesBox.get_children();
        if (this.categorysHidden && !hide) {
            for (var i = 1; i < categoriesBoxChilren.length; i++) {
                let item_actor = categoriesBoxChilren[i];
                item_actor.show();
                if(i == categoriesBoxChilren.length-1)
                    item_actor.hide();
            }
            this.categorysHidden = false;
            this._allAppsLabel.set_text(_("All Applications"));
        } else if (hide) {
            for (var i = 1; i < categoriesBoxChilren.length; i++) {
                let item_actor = categoriesBoxChilren[i];
                item_actor.hide();
                if(i == categoriesBoxChilren.length-1)
                    item_actor.show();
            }
            this.categorysHidden = true;
            this._allAppsLabel.set_text(_("Search results"));
        }

    },
    _toggleView: function(isSearchView){
        if(isSearchView){
            this.top.set_child(this.searchView, {
                x_align: St.Align.END,
                x_fill: true
            });
            this.shutdownBox.add_style_class_name('light');
            this.shutdownButton.shutdown.add_style_class_name('dark');
            this.shutdownButton.shutdownMenu.actor.add_style_class_name('dark');
        }else{
            this.top.set_child(this.defaultView, {
                x_align: St.Align.END,
                x_fill: true
            });
            this.shutdownBox.remove_style_class_name('light');
            this.shutdownButton.shutdown.remove_style_class_name('dark');
            this.shutdownButton.shutdownMenu.actor.remove_style_class_name('dark');
        }
    },

    reset: function(){
        this.favsBox.show();
        let children = this.favsBox.get_children();
        children.forEach(function(child){
            child.emit("style_changed");
        });
        this.categoriesScrollBox.hide();
        this.allProButton.reset();
        let actors = this.categoriesBox.get_children();
        for (let i = 0; i < actors.length; i++) {
            let actor = actors[i];
            if (actor._delegate && actor._delegate.menu){
                //actor._delegate.removeItems();
                actor._delegate.menu.closex(); 
            }
        }
        this._refreshFavs();
        this._refreshRecent();
    },

    resetSearch: function() {
        this.searchEntry.set_text("");
        this.searchActive = false;
        global.stage.set_key_focus(this.searchEntry);
    },

    _onSearchTextChanged: function(se, prop) {
        this._clearSelections(this.categoriesBox);
        this.searchActive = this.searchEntry.get_text() != '';
        if (this.searchActive) {
            this.searchEntry.set_secondary_icon(this._searchActiveIcon);

            if (this._searchIconClickedId == 0) {
                this._searchIconClickedId = this.searchEntry.connect('secondary-icon-clicked', Lang.bind(this, function() {
                    this.resetSearch();
                }));
            }
        } else {
            if (this._searchIconClickedId > 0) this.searchEntry.disconnect(this._searchIconClickedId);
            this._searchIconClickedId = 0;

            this.searchEntry.set_secondary_icon(this._searchInactiveIcon);
        }
        if (!this.searchActive) {
            if (this._searchTimeoutId > 0) {
                Mainloop.source_remove(this._searchTimeoutId);
                this._searchTimeoutId = 0;
            }
            //this._categoryButtonsHidden(false);
            this._toggleView(false);
            return;
        }
        if (this._searchTimeoutId > 0) return;
        this._searchTimeoutId = Mainloop.timeout_add(150, Lang.bind(this, this._doSearch));
    },

    _listBookmarks: function(pattern) {
        let bookmarks = Main.placesManager.getBookmarks();
        var res = new Array();
        for (let id = 0; id < bookmarks.length; id++) {
            if (!pattern || bookmarks[id].name.toLowerCase().indexOf(pattern) != -1) res.push(bookmarks[id]);
        }
        return res;
    },

    _listDevices: function(pattern) {
        let devices = Main.placesManager.getMounts();
        var res = new Array();
        for (let id = 0; id < devices.length; id++) {
            if (!pattern || devices[id].name.toLowerCase().indexOf(pattern) != -1) res.push(devices[id]);
        }
        return res;
    },

    _listApplications: function(category_menu_id, pattern) {
        var applist;
        if (category_menu_id) applist = this.applicationsByCategory[category_menu_id];
        else {
            applist = new Array();
            for (directory in this.applicationsByCategory) applist = applist.concat(this.applicationsByCategory[directory]);
        }

        var res;
        if (pattern) {
            res = new Array();
            for (var i in applist) {
                let app = applist[i];
                if (app.get_name().toLowerCase().indexOf(pattern) != -1 || (app.get_description() && app.get_description().toLowerCase().indexOf(pattern) != -1) || (app.get_id() && app.get_id().slice(0, -8).toLowerCase().indexOf(pattern) != -1)) res.push(app);
            }
        } else res = applist;

        res.sort(function(a, b) {
            return a.get_name().toLowerCase() > b.get_name().toLowerCase();
        });

        return res;
    },

    _listApplicationsGroupByCategory: function(pattern) {
        var res = {};
        var applist;
        var arr;
        var app;
        for (var directory in this.applicationsByCategory) {
            applist = this.applicationsByCategory[directory];
            arr = new Array();
            for (let i = 0;  i < applist.length; i++) {
                app = applist[i];
                //if ((app.get_name().toLowerCase().indexOf(pattern) != -1) || (app.get_description() && app.get_description().toLowerCase().indexOf(pattern) != -1) || (app.get_id() && app.get_id().slice(0, -8).toLowerCase().indexOf(pattern) != -1))
                //    arr.push(app);
                //if ((app.get_name().toLowerCase().indexOf(pattern) != -1) || (app.get_description() && app.get_description().toLowerCase().indexOf(pattern) != -1))
                
                let _name = app.get_name();
                let _pinyin = pinyin.getFullChars(app.get_name()).toLowerCase();
                let _cname = pinyin.getCamelChars(app.get_name()).toLowerCase();
                if(_name.indexOf(pattern) >=0 || _pinyin.indexOf(pattern) >= 0 || _cname.indexOf(pattern) >= 0){
                    arr.push(app);
                }
            }
            if(arr.length){
                res[directory] = arr;
            }
        }
        return res;
    },

    _doSearch: function() {

        this._toggleView(true);

        this._searchTimeoutId = 0;
        let pattern = this.searchEntryText.get_text().replace(/^\s+/g, '').replace(/\s+$/g, '').toLowerCase();
        if (pattern == this._previousSearchPattern) return false;
        this._previousSearchPattern = pattern;

        this._activeContainer = null;
        this._selectedItemIndex = null;
        this._previousSelectedItemIndex = null;
        
        // _listApplications returns all the applications when the search
        // string is zero length. This will happend if you type a space
        // in the search entry.
        this.searchResultBox.destroy_children();
        let vbox = new St.BoxLayout({vertical: true});

        let moreBox = new St.BoxLayout({vertical: true});
        let _button = new SearchResultButton(this, null);
        let _icon = new St.Icon({
            icon_name: 'search',
            icon_size: 18,
            icon_type: St.IconType.FULLCOLOR
        });
        let _label = new St.Label({
            text: 'See more results',
            style_class: 'menu-application-button-label'
        });
        _button.addActor(_icon);
        _button.addActor(_label);
        //moreBox.add(_button.actor);
        
        this.searchResultBox.add(vbox);
        this.searchResultBox.add(moreBox);
        vbox.height = this.searchResultBox.height - 30;
        //var appResults = this._listApplications(null, pattern);
        var appResults = this._listApplicationsGroupByCategory(pattern);

        var placesResults = new Array();

        var bookmarks = this._listBookmarks(pattern);
        for (var i in bookmarks) placesResults.push(bookmarks[i]);

        var devices = this._listDevices(pattern);
        for (var i in devices) placesResults.push(devices[i]);

        if(JSON.stringify(appResults) === "{}") {
            let label = new St.Label({text: _('No match for search terms')});
            let bin = new St.Bin({
                style_class: 'menu-search-result-label'
            });
            bin.set_child(label);
            vbox.add(bin);
            return false;
        }

        let maxHeight = vbox.height - 30;
        let labelHeight = 24;
        let separatorHeight = 9;
        let buttonHeight = 26;
        let naturalHeight = 0;

        for(var i in appResults){
            let applist = appResults[i];
            let label = new St.Label({text: this.categoryMap[i], style_class: 'favorites-label'});

            if(applist && applist.length){
                naturalHeight += (labelHeight + separatorHeight);
                if(naturalHeight >= maxHeight)
                    break;
                vbox.add(label);
                vbox.add(new Separator({style_class: 'menu-favorites-separator'}).actor);
                for(let i = 0; i < applist.length; i++){
                    naturalHeight += buttonHeight;
                    if(naturalHeight >= maxHeight)
                        break;
                    let button = new SearchResultButton(this, applist[i]);
                    vbox.add(button.actor);
                }
            }
        }

        let categoriesBoxChilren = this.categoriesBox.get_children();
        if (categoriesBoxChilren.length > 0) {
            this._activeContainer = this.categoriesBox;
            this._selectedItemIndex = 0;
            let item_actor = categoriesBoxChilren[this._selectedItemIndex];
            if (item_actor && item_actor !== this.searchEntry && item_actor._delegate) {
                item_actor._delegate.emit('enter-event');
            }
        }
        //this._categoryButtonsHidden(true);
        return false;
    }


};

function main(metadata, orientation, panel_height, instance_id) {
    let myApplet = new MyApplet(orientation, panel_height, instance_id);
    return myApplet;
}
